import numpy as np
import pandas as pd
import pyomo.environ as pyo

from ..model_construction.utilities import determine_network_energy_consumption


def _determine_carriers_from_time_series(time_series: pd.DataFrame) -> list:
    """
    Determines carriers that are used in time_series

    :param pd.Dataframe time_series: Time series
    """
    carriers = []
    for car in time_series.columns.get_level_values("Carrier"):
        if np.any(time_series[car]):
            carriers.append(car)

    return list(set(carriers))


def _determine_carriers_from_technologies(technology_data: dict) -> list:
    """
    Determines carriers that are used for technologies

    :param dict technology_data: Dict with technology data
    :return: list of carriers used in technologies
    """
    carriers = []
    for tec in technology_data:
        input_carriers = technology_data[tec].component_options.input_carrier
        carriers.extend(input_carriers)
        output_carriers = technology_data[tec].component_options.output_carrier
        carriers.extend(output_carriers)

    return list(set(carriers))


def _determine_carriers_from_networks(network_data) -> list:
    """
    Determines carriers that are used for networks

    :param dict network_data: Dict with network data
    :return: list of carriers used in networks
    """
    carriers = []
    for netw in network_data:
        # Todo: This can be further extended to check if node is connected to network
        carriers.extend([network_data[netw].component_options.transported_carrier])

        if network_data[netw].component_options.energyconsumption:
            carriers.extend(network_data[netw].energy_consumption.keys())

    return list(set(carriers))


def construct_node_block(b_node, data: dict, set_t_full, set_t_clustered):
    """
    Adds all nodes with respective data to the model

    This function initializes parameters and decision variables for all considered
    nodes.

    **Set declarations:**

    - set_technologies: Set for all technologies at respective node
    - set_carriers: Set of carriers used at node (this is a subset of all carriers)

    **Parameter declarations:**

    - para_demand: Demand for each time step and carrier
    - para_production_profile: Maximal generic production profile for each time step
      and carrier
    - para_import_price: Import Prices for each time step and carrier
    - para_export_price: Export Prices for each time step and carrier
    - para_import_limit: Import Limits for each time step and carrier
    - para_export_limit: Export Limits for each time step and carrier
    - para_import_emissionfactors: Emission factors of imports for each time step and
      carrier
    - para_export_emissionfactors: Emission factors of exports for each time step and
      carrier
    - para_carbon_subsidy: Carbon subsidy for negative emissions
    - para_carbon_tax: Carbon tax for positive emissions

    **Variable declarations:**

    - var_import_flow: Import Flow for each time step and carrier
    - var_export_flow: Export Flow for each time step and carrier
    - var_netw_inflow: Network Inflow for each time step and carrier
    - var_netw_outflow: Network Outflow for each time step and carrier
    - var_netw_consumption: Network consumption (if present)
    - var_generic_production: Actual generic production
    - var_import_emissions_pos: Positive emissions from imports
    - var_import_emissions_neg: Negative emissions from imports
    - var_export_emissions_pos: Positive emissions from exports
    - var_export_emissions_neg: Negative emissions from exports
    - var_car_emissions_pos: Sum of positive emissions
    - var_car_emissions_neg: Sum of negative emissions

    **Constraint declarations**

    - Generic production: Equal to the parameter if curtailment is not possible,
      otherwise less-or-equal
    - Calculate import emissions (positive and negative): emissions = import *
      emissions-factor
    - Calculate export emissions (positive and negative): emissions = export *
      emissions-factor
    - Calculate carrier emissions as a sum of positive/negative import and export
      emissions

    :param b_node: pyomo block with node model
    :param dict data: data containing model configuration
    :param set_t_full: pyomo set containing full resolution timesteps
    :param set_t_clustered: pyomo set containing clustered resolution timesteps
    :return: pyomo block with node model
    """

    # PREPROCESSING
    # Collect data for node and period
    config = data["config"]

    # Determine carriers used at node
    carriers = []
    carriers.extend(
        _determine_carriers_from_time_series(data["time_series"]["CarrierData"])
    )
    carriers.extend(_determine_carriers_from_technologies(data["technology_data"]))
    if not config["energybalance"]["copperplate"]["value"]:
        carriers.extend(_determine_carriers_from_networks(data["network_data"]))
    carriers = list(set(carriers))

    # SETS
    b_node.set_technologies = pyo.Set(initialize=list(data["technology_data"].keys()))
    b_node.set_carriers = pyo.Set(initialize=list(set(carriers)))

    # Time aggregation
    config = data["config"]
    if config["optimization"]["typicaldays"]["N"]["value"] == 0:
        set_t = set_t_full
    elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
        set_t = set_t_clustered
    elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
        set_t = set_t_full

    # PARAMETERS
    def create_carrier_parameter(key, par_mutable=False):
        # Convert to dict/list for performance
        ts = {}
        for car in b_node.set_carriers:
            ts[car] = {}
            ts[car][key] = data["time_series"]["CarrierData"][car][key].to_list()

        def init_carrier_parameter(para, t, car):
            """Rule initiating a carrier parameter"""
            return ts[car][key][t - 1]

        parameter = pyo.Param(
            set_t, b_node.set_carriers, rule=init_carrier_parameter, mutable=par_mutable
        )
        return parameter

    def create_carbonprice_parameter(key):
        # Convert to dict/list for performance
        ts = data["time_series"]["CarbonCost"]["global"][key].to_list()

        def init_carbonprice_parameter(para, t):
            """Rule initiating a carrier parameter"""
            return ts[t - 1]

        parameter = pyo.Param(set_t, rule=init_carbonprice_parameter, mutable=False)
        return parameter

    if config["optimization"]["monte_carlo"]["N"]["value"] != 0:
        par_mutable = True
    else:
        par_mutable = False

    b_node.para_demand = create_carrier_parameter("Demand")
    b_node.para_production_profile = create_carrier_parameter("Generic production")
    b_node.para_import_price = create_carrier_parameter(
        "Import price", par_mutable=par_mutable
    )
    b_node.para_export_price = create_carrier_parameter(
        "Export price", par_mutable=par_mutable
    )
    b_node.para_import_limit = create_carrier_parameter("Import limit")
    b_node.para_export_limit = create_carrier_parameter("Export limit")
    b_node.para_import_emissionfactors = create_carrier_parameter(
        "Import emission factor"
    )
    b_node.para_export_emissionfactors = create_carrier_parameter(
        "Export emission factor"
    )
    b_node.para_carbon_subsidy = create_carbonprice_parameter("subsidy")
    b_node.para_carbon_tax = create_carbonprice_parameter("price")

    # VARIABLES
    def init_import_bounds(var, t, car):
        return (0, b_node.para_import_limit[t, car])

    b_node.var_import_flow = pyo.Var(
        set_t, b_node.set_carriers, bounds=init_import_bounds
    )

    def init_export_bounds(var, t, car):
        return (0, b_node.para_export_limit[t, car])

    b_node.var_export_flow = pyo.Var(
        set_t, b_node.set_carriers, bounds=init_export_bounds
    )

    if not config["energybalance"]["copperplate"]["value"]:
        b_node.var_netw_inflow = pyo.Var(set_t, b_node.set_carriers)
        b_node.var_netw_outflow = pyo.Var(set_t, b_node.set_carriers)
        network_energy_consumption = determine_network_energy_consumption(
            data["network_data"]
        )
        if network_energy_consumption:
            b_node.var_netw_consumption = pyo.Var(set_t, b_node.set_carriers)

    # Generic production profile
    b_node.var_generic_production = pyo.Var(
        set_t, b_node.set_carriers, within=pyo.NonNegativeReals
    )

    # Emissions
    b_node.var_import_emissions_pos = pyo.Var(set_t, b_node.set_carriers)
    b_node.var_import_emissions_neg = pyo.Var(set_t, b_node.set_carriers)
    b_node.var_export_emissions_pos = pyo.Var(set_t, b_node.set_carriers)
    b_node.var_export_emissions_neg = pyo.Var(set_t, b_node.set_carriers)
    b_node.var_car_emissions_pos = pyo.Var(set_t, within=pyo.NonNegativeReals)
    b_node.var_car_emissions_neg = pyo.Var(set_t, within=pyo.NonNegativeReals)

    # CONSTRAINTS
    # Generic production constraint
    def init_generic_production(const, t, car):
        if data["energybalance_options"][car]["curtailment_possible"] == 0:
            return (
                b_node.para_production_profile[t, car]
                == b_node.var_generic_production[t, car]
            )
        elif data["energybalance_options"][car]["curtailment_possible"] == 1:
            return (
                b_node.para_production_profile[t, car]
                >= b_node.var_generic_production[t, car]
            )

    b_node.const_generic_production = pyo.Constraint(
        set_t, b_node.set_carriers, rule=init_generic_production
    )

    # Emission constraints
    def init_import_emissions_pos(const, t, car):
        if b_node.para_import_emissionfactors[t, car] >= 0:
            return (
                b_node.var_import_flow[t, car]
                * b_node.para_import_emissionfactors[t, car]
                == b_node.var_import_emissions_pos[t, car]
            )
        else:
            return 0 == b_node.var_import_emissions_pos[t, car]

    b_node.const_import_emissions_pos = pyo.Constraint(
        set_t, b_node.set_carriers, rule=init_import_emissions_pos
    )

    def init_export_emissions_pos(const, t, car):
        if b_node.para_export_emissionfactors[t, car] >= 0:
            return (
                b_node.var_export_flow[t, car]
                * b_node.para_export_emissionfactors[t, car]
                == b_node.var_export_emissions_pos[t, car]
            )
        else:
            return 0 == b_node.var_export_emissions_pos[t, car]

    b_node.const_export_emissions_pos = pyo.Constraint(
        set_t, b_node.set_carriers, rule=init_export_emissions_pos
    )

    def init_import_emissions_neg(const, t, car):
        if b_node.para_import_emissionfactors[t, car] < 0:
            return (
                b_node.var_import_flow[t, car]
                * (-b_node.para_import_emissionfactors[t, car])
                == b_node.var_import_emissions_neg[t, car]
            )
        else:
            return 0 == b_node.var_import_emissions_neg[t, car]

    b_node.const_import_emissions_neg = pyo.Constraint(
        set_t, b_node.set_carriers, rule=init_import_emissions_neg
    )

    def init_export_emissions_neg(const, t, car):
        if b_node.para_export_emissionfactors[t, car] < 0:
            return (
                b_node.var_export_flow[t, car]
                * (-b_node.para_export_emissionfactors[t, car])
                == b_node.var_export_emissions_neg[t, car]
            )
        else:
            return 0 == b_node.var_export_emissions_neg[t, car]

    b_node.const_export_emissions_neg = pyo.Constraint(
        set_t, b_node.set_carriers, rule=init_export_emissions_neg
    )

    def init_car_emissions_pos(const, t):
        return (
            sum(
                b_node.var_import_emissions_pos[t, car]
                + b_node.var_export_emissions_pos[t, car]
                for car in b_node.set_carriers
            )
            == b_node.var_car_emissions_pos[t]
        )

    b_node.const_car_emissions_pos = pyo.Constraint(set_t, rule=init_car_emissions_pos)

    def init_car_emissions_neg(const, t):
        return (
            sum(
                b_node.var_import_emissions_neg[t, car]
                + b_node.var_export_emissions_neg[t, car]
                for car in b_node.set_carriers
            )
            == b_node.var_car_emissions_neg[t]
        )

    b_node.const_car_emissions_neg = pyo.Constraint(set_t, rule=init_car_emissions_neg)

    return b_node



from .template_creation import (
    create_input_data_folder_template,
    create_optimization_templates,
    initialize_configuration_templates,
    initialize_topology_templates,
    create_montecarlo_template_csv,
    show_available_networks,
    show_available_technologies,
)
from .data_loading import (
    copy_network_data,
    copy_technology_data,
    fill_carrier_data,
    load_climate_data_from_api,
)



import pandas as pd

from ...component import ModelComponent


def fit_ccs_coeff(co2_concentration: float, ccs_data: dict, climate_data: pd.DataFrame):
    """
    Obtain bounds and input ratios for CCS

    Calculates the amount of input (and their bounds) required for each unit of CO2 entering the carbon capture (CC)
    object. The minimum and maximum size parameters are multiplied by the CO2 concentration, so that the units
    of the size becomes t/h of CO2 in. These are also the units used for the rest of the model.
    So far, only post-combustion MEA has been modelled (based on Eq. 7 in Weimann et Al. (2023), A thermodynamic-based
    mixed-integer linear model of post-combustion carbon capture for reliable use in energy system optimisation
    https://doi.org/10.1016/j.apenergy.2023.120738).

    :param float co2_concentration: CO2 concentration for CCS
    :param dict ccs_coeff: data of the CCS technology
    :param pd.Dataframe climate_data: dataframe containing climate data
    :return: CCS data updated with the bounds and input ratios for CCS
    """
    # Recalculate unit_capex
    molar_mass_CO2 = 44.01
    # convert kmol/s of fluegas to ton/h of CO2molar_mass_CO2 = 44.01
    convert2t_per_h = molar_mass_CO2 * co2_concentration * 3.6

    capture_rate = ccs_data["Performance"]["capture_rate"]
    ccs_data["Economics"]["unit_CAPEX"] = (
        (
            ccs_data["Economics"]["CAPEX_kappa"] / convert2t_per_h
            + ccs_data["Economics"]["CAPEX_lambda"]
        )
        * capture_rate
        * co2_concentration
    ) / convert2t_per_h

    ccs_data["Economics"]["fix_CAPEX"] = ccs_data["Economics"]["CAPEX_zeta"]

    ccs_data = ModelComponent(ccs_data)

    molar_mass_CO2 = 44.01

    # Recalculate min/max size to have it in t/hCO2_in
    ccs_data.input_parameters.size_min = (
        ccs_data.input_parameters.size_min * co2_concentration
    )
    ccs_data.input_parameters.size_max = (
        ccs_data.input_parameters.size_max * co2_concentration
    )

    # Calculate input ratios
    ccs_data.processed_coeff.time_independent["size_min"] = (
        ccs_data.input_parameters.size_min
    )
    ccs_data.processed_coeff.time_independent["size_max"] = (
        ccs_data.input_parameters.size_max
    )
    ccs_data.processed_coeff.time_independent["capture_rate"] = capture_rate
    if "MEA" in ccs_data.component_options.technology_model:
        input_ratios = {}
        for car in ccs_data.component_options.input_carrier:
            input_ratios[car] = (
                ccs_data.input_parameters.performance_data["eta"][car]
                + ccs_data.input_parameters.performance_data["omega"][car]
                * co2_concentration
            ) / (co2_concentration * molar_mass_CO2 * 3.6)
        ccs_data.processed_coeff.time_independent["input_ratios"] = input_ratios
    else:
        raise Exception(
            "Only CCS type MEA is modelled so far. ccs_type in the json file of the "
            "technology must include MEA"
        )

    return ccs_data



import random
import warnings
from pathlib import Path
import pyomo.environ as pyo
import os
import time
import numpy as np
import pandas as pd
import sys
import datetime

from .utilities import get_set_t
from .data_management import DataHandle, read_tec_data
from .model_construction import *
from .result_management.read_results import add_values_to_summary
from .utilities import get_glpk_parameters, get_gurobi_parameters
from .result_management import *
from .components.utilities import (
    annualize,
    set_discount_rate,
    perform_disjunct_relaxation,
)
import logging

log = logging.getLogger(__name__)


class ModelHub:
    """
    Class to construct and manipulate an energy system model.

    When constructing an instance, it reads data to the instance and initializes all
    attributes of the ModelHub class:

    - self.data: Data container
    - self.model: Model container
    - self.solution: Solution container
    - self.solver: Solver container
    - self.last_solve_info: Information on last solution that is written to the summary(
      pareto point, time stage,...)
    - self.info_pareto: Current pareto point (if used)
    - self.info_solving_algorithms: Information on time aggregation algorithms
    - self.info_monte_carlo: Information on monte carlo runs
    """

    def __init__(self):
        """
        Constructor
        """
        self.data = DataHandle()
        self.model = {}
        self.solution = {}
        self.solver = None
        self.last_solve_info = {}
        self.info_pareto = {}
        self.info_pareto["pareto_point"] = -1
        self.info_solving_algorithms = {}
        self.info_solving_algorithms["aggregation_model"] = "Full"
        self.info_solving_algorithms["aggregation_data"] = "Full"
        self.info_solving_algorithms["time_stage"] = 1
        self.info_monte_carlo = {}
        self.info_monte_carlo["monte_carlo_run"] = -1

    def read_data(
        self, data_path: Path | str, start_period: int = None, end_period: int = None
    ):
        """
        Reads in data from the specified path. The data is specified as the DataHandle
        class. Specifying the start_period and end_period parameter allows to run a
        time horizon than as specified in the topology (e.g. for testing)

        :param Path, str data_path: Path of folder structure to read data from
        :param int start_period: starting period of the model
        :param int end_period: end period of the model
        """
        log_msg = "--- Reading in data ---"
        log.info(log_msg)
        self.data.set_settings(data_path, start_period, end_period)
        self.data.read_data()
        self._perform_preprocessing_checks()

        log_msg = "--- Reading in data complete ---"
        log.info(log_msg)

    def _perform_preprocessing_checks(self):
        """
        Checks consistency of input data, before constructing or solving the model

        - Save path must exist
        - monte carlo and pareto cannot be used at the same time
        - dynamics checks
        :return:
        """
        config = self.data.model_config
        topology = self.data.topology

        # Is solver available?
        try:
            mock_model = pyo.ConcreteModel()
            if config["solveroptions"]["solver"]["value"] == "gurobi":
                solver = get_gurobi_parameters(config["solveroptions"])
            elif config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
                solver = get_gurobi_parameters(config["solveroptions"])
                self.solver.set_instance(mock_model)
            elif config["solveroptions"]["solver"]["value"] == "glpk":
                solver = get_glpk_parameters(config["solveroptions"])
            solver.solve(mock_model)
        except:
            raise Exception(
                "The solver you are trying to use is not available. This "
                "could be due to the following two reasons: (1) it is not "
                "implemented (currently AdOpT-NET0 supports gurobi and "
                "glpk. (2) If you are using gurobi, make sure that the "
                "gurobipy version installed matches the gurobi version "
                "installed."
            )

        # Check if node locations are specified
        if self.data.node_locations.isnull().values.any():
            raise Exception(
                "Please specify longitude, latitude and altitude for each node"
            )

        # Check if save-path exists
        save_path = Path(config["reporting"]["save_path"]["value"])
        if not os.path.exists(save_path) or not os.path.isdir(save_path):
            raise FileNotFoundError(
                f"The folder you want to save your results to ('{save_path}') does not "
                f"exist. Create the folder or change the folder "
                f"name in the ModelConfig"
            )

        # Monte carlo and pareto
        if (config["optimization"]["objective"]["value"] == "pareto") and (
            config["optimization"]["monte_carlo"]["N"]["value"] > 0
        ):
            raise Exception("Monte carlo and pareto is not allowed at the same time")

        # Dynamics and time aggregation algorithms
        if config["optimization"]["typicaldays"]["N"]["value"] != 0:
            if config["performance"]["dynamics"]["value"]:
                raise Exception(
                    "Dynamics and clustering with typical days is not " "allowed"
                )
            for period in topology["investment_periods"]:
                for node in topology["nodes"]:
                    for tec_name in self.data.technology_data[period][node]:
                        tec = self.data.technology_data[period][node][tec_name]
                        if ("ramping_const_int" in tec.processed_coeff.dynamics) and (
                            tec.processed_coeff.dynamics["ramping_const_int"] != -1
                        ):
                            raise Exception(
                                f"Ramping constraint with integers (ramping_const_int) for technology {tec_name} "
                                f"needs to be -1 when clustering with typical days"
                            )

        if config["optimization"]["timestaging"]["value"] != 0:
            if config["performance"]["dynamics"]["value"]:
                raise Exception(
                    "Dynamics and two-stage averaging algorithm is not " "allowed"
                )
            for period in topology["investment_periods"]:
                for node in topology["nodes"]:
                    for tec_name in self.data.technology_data[period][node]:
                        tec = self.data.technology_data[period][node][tec_name]
                        if ("ramping_time" in tec.processed_coeff.dynamics) and (
                            tec.processed_coeff.dynamics["ramping_time"] != -1
                        ):
                            raise Exception(
                                f"Ramping Rate for technology {tec_name} "
                                f"needs to be -1 when two-stage averaging algorithm is used"
                            )

        # check if technologies have dynamic parameters
        if config["performance"]["dynamics"]["value"]:
            for node in self.data.topology.nodes:
                for tec in self.data.technology_data[node]:
                    if self.data.technology_data[node][tec].technology_model in [
                        "CONV1",
                        "CONV2",
                        "CONV3",
                    ]:
                        par_check = [
                            "max_startups",
                            "min_uptime",
                            "min_downtime",
                            "SU_load",
                            "SD_load",
                            "SU_time",
                            "SD_time",
                        ]
                        for par in par_check:
                            if (
                                par
                                not in self.data.technology_data[node][
                                    tec
                                ].processed_coeff.dynamics
                            ):
                                raise ValueError(
                                    f"The technology '{tec}' does not have dynamic parameter '{par}'. Add the parameters in the "
                                    f"json files or switch off the dynamics."
                                )

    def construct_model(self):
        """
        Constructs the model. The model structure is as follows:

        **Global sets**

        - set_periods: set of investment periods
        - set_nodes: set of nodes
        - set_carriers: set of carriers modelled

        **Global variables**

        - var_npv: net present value of all costs
        - var_emissions_net: net emissions over all investment periods

        **Rest of model**
        The rest of the model is organized in nested, hierarchical pyomo modelling blocks:

        Investment Period Block

            Network Block

            Node Block

                Technology Block
        """
        log_msg = "--- Constructing Model ---"
        log.info(log_msg)
        start = time.time()

        # Determine aggregation
        config = self.data.model_config

        # Clustered data
        self.info_solving_algorithms["aggregation_model"] = "full"
        self.info_solving_algorithms["aggregation_data"] = "full"

        if config["optimization"]["typicaldays"]["N"]["value"] != 0:
            if config["optimization"]["typicaldays"]["method"]["value"] == 1:
                self.info_solving_algorithms["aggregation_model"] = "clustered"
                self.info_solving_algorithms["aggregation_data"] = "clustered"
            elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
                self.info_solving_algorithms["aggregation_model"] = "clustered"
                self.info_solving_algorithms["aggregation_data"] = "full"
            else:
                raise Exception("clustering method needs to be 1 or 2")

        # Averaged data
        if config["optimization"]["timestaging"]["value"] != 0:
            if self.info_solving_algorithms["time_stage"] == 1:
                self.info_solving_algorithms["aggregation_model"] = "averaged"
                self.info_solving_algorithms["aggregation_data"] = "averaged"

        # INITIALIZE MODEL
        aggregation_model = self.info_solving_algorithms["aggregation_model"]
        aggregation_data = self.info_solving_algorithms["aggregation_data"]
        self.model[aggregation_model] = pyo.ConcreteModel()

        # GET DATA
        model = self.model[aggregation_model]
        topology = self.data.topology
        config = self.data.model_config

        # DEFINE GLOBAL SETS
        # Nodes, Carriers, Technologies, Networks
        model.set_periods = pyo.Set(initialize=topology["investment_periods"])
        model.set_nodes = pyo.Set(initialize=topology["nodes"])
        model.set_carriers = pyo.Set(initialize=topology["carriers"])

        # DEFINE GLOBAL VARIABLES
        model.var_npv = pyo.Var()
        model.var_emissions_net = pyo.Var()

        # INVESTMENT PERIOD BLOCK
        def init_period_block(b_period):
            """Pyomo rule to initialize a block holding all investment periods"""

            # Get data for investment period
            investment_period = b_period.index()
            data_period = get_data_for_investment_period(
                self.data, investment_period, aggregation_data
            )
            # Add sets, parameters, variables, constraints to block
            b_period = construct_investment_period_block(b_period, data_period)

            # NETWORK BLOCK
            if not config["energybalance"]["copperplate"]["value"]:

                def init_network_block(b_netw, netw):
                    """Pyomo rule to initialize a block holding all networks"""
                    # Add sets, parameters, variables, constraints to block
                    b_netw = construct_network_block(
                        b_netw,
                        data_period,
                        model.set_nodes,
                        b_period.set_t_full,
                        b_period.set_t_clustered,
                    )

                    return b_netw

                b_period.network_block = pyo.Block(
                    b_period.set_networks, rule=init_network_block
                )

            # NODE BLOCK
            def init_node_block(b_node, node):
                """Pyomo rule to initialize a block holding all nodes"""
                # Get data for node
                data_node = get_data_for_node(data_period, node)

                # Add sets, parameters, variables, constraints to block
                b_node = construct_node_block(
                    b_node, data_node, b_period.set_t_full, b_period.set_t_clustered
                )

                # TECHNOLOGY BLOCK
                def init_technology_block(b_tec, tec):
                    b_tec = construct_technology_block(
                        b_tec, data_node, b_period.set_t_full, b_period.set_t_clustered
                    )

                    return b_tec

                b_node.tech_blocks_active = pyo.Block(
                    b_node.set_technologies, rule=init_technology_block
                )

                return b_node

            b_period.node_blocks = pyo.Block(model.set_nodes, rule=init_node_block)

            return b_period

        model.periods = pyo.Block(model.set_periods, rule=init_period_block)

        log_msg = f"Constructing model completed in {str(round(time.time() - start))}s"
        log.info(log_msg)

    def construct_balances(self):
        """
        Constructs the energy balance, emission balance and calculates costs
        """
        log_msg = "Constructing balances..."
        log.info(log_msg)
        start = time.time()

        config = self.data.model_config
        data = self.data
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        model = delete_all_balances(model)

        if not config["energybalance"]["copperplate"]["value"]:
            model = construct_network_constraints(model, config)
            model = construct_nodal_energybalance(model, config)
        else:
            model = construct_global_energybalance(model, config)

        model = construct_emission_balance(model, data)
        model = construct_system_cost(model, data)
        model = construct_global_balance(model)

        log_msg = (
            f"Constructing balances completed in {str(round(time.time() - start))}s"
        )
        log.warning(log_msg)

    def solve(self):
        """
        Defines objective and solves model
        """
        config = self.data.model_config

        objective = config["optimization"]["objective"]["value"]

        self._define_solver_settings()

        if config["optimization"]["monte_carlo"]["N"]["value"]:
            self._solve_monte_carlo(objective)
        elif objective == "pareto":
            self._solve_pareto()
        else:
            self._optimize(objective)

    def quick_solve(self):
        """
        Quick-solves the model (constructs model and balances and solves model).

        This method lumbs together the following functions for convenience:
        - :func:`~adopt_net0.modelhub.construct_model`
        - :func:`~adopt_net0.modelhub.construct_balances`
        - :func:`~adopt_net0.modelhub.solve`
        """
        self.construct_model()
        self.construct_balances()
        self.solve()

    def write_results(self):
        """
        Writes optimization results of a model run to folder
        """
        # Write H5 File

        solution_available = True
        if self.solution.solver.termination_condition in [
            pyo.TerminationCondition.infeasibleOrUnbounded,
            pyo.TerminationCondition.infeasible,
            pyo.TerminationCondition.unbounded,
        ]:
            solution_available = False

        if solution_available:

            config = self.data.model_config

            save_summary_path = Path.joinpath(
                Path(config["reporting"]["save_summary_path"]["value"]), "Summary.xlsx"
            )

            model_info = self.last_solve_info

            model = self.model[self.info_solving_algorithms["aggregation_model"]]

            summary_dict = write_optimization_results_to_h5(
                model, self.solution, model_info, self.data
            )

            # Write Summary
            if not os.path.exists(save_summary_path):
                summary_df = pd.DataFrame(data=summary_dict, index=[0])
                summary_df.to_excel(
                    save_summary_path, index=False, sheet_name="Summary"
                )
            else:
                summary_existing = pd.read_excel(save_summary_path)
                pd.concat(
                    [summary_existing, pd.DataFrame(data=summary_dict, index=[0])]
                ).to_excel(save_summary_path, index=False, sheet_name="Summary")

    def add_technology(self, investment_period: str, node: str, technologies: list):
        """
        Adds technologies retrospectively to the model.

        After adding a technology to a node, all balances need to be re-constructed,
        To solve the model again run, :func:`~construct_balances` and then
        :func:`~solve`.

        :param str investment_period: name of investment period for which technology is added
        :param str node: name of node for which technology is added
        :param list technologies: list of technologies that should be added
        :return None:
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        # Make sure that no aggregation algorithm is used
        config = self.data.model_config
        if (config["optimization"]["typicaldays"]["N"]["value"] != 0) or (
            config["optimization"]["timestaging"]["value"] != 0
        ):
            raise Exception(
                "You cannot add a technolgy retrospectively if using time aggragation algorithms"
            )

        # Read technology data
        data_node = {
            "technology_data": {},
            "config": config,
            "topology": self.data.topology,
        }
        for technology in technologies:
            # read in technology data
            tec_data = read_tec_data(
                technology,
                self.data.data_path
                / investment_period
                / "node_data"
                / node
                / "technology_data",
            )
            # fit technology data
            tec_data.fit_technology_performance(
                self.data.time_series["full"][investment_period][node]["ClimateData"][
                    "global"
                ],
                self.data.node_locations.loc[node, :],
            )
            # add technology data to data handle
            self.data.technology_data[investment_period][node][technology] = tec_data
            data_node["technology_data"][technology] = tec_data

        # Add technology to node
        b_period = model.periods[investment_period]
        b_node = b_period.node_blocks[node]

        # Create new technology block containing all new technologies
        def init_technology_block(b_tec, tec):
            b_tec = construct_technology_block(
                b_tec, data_node, b_period.set_t_full, b_period.set_t_clustered
            )

            return b_tec

        b_node.tech_blocks_new = pyo.Block(technologies, rule=init_technology_block)

        # If it exists, carry over active tech blocks to temporary block
        if b_node.find_component("tech_blocks_active"):
            b_node.tech_blocks_existing = pyo.Block(b_node.set_technologies)
            for tec in b_node.set_technologies:
                b_node.tech_blocks_existing[tec].transfer_attributes_from(
                    b_node.tech_blocks_active[tec]
                )
            b_node.del_component(b_node.tech_blocks_active)
        if b_node.find_component("tech_blocks_active_index"):
            b_node.del_component(b_node.tech_blocks_active_index)

        # Create a block containing all active technologies at node
        if not set(technologies).issubset(b_node.set_technologies):
            b_node.set_technologies.add(technologies)

        def init_active_technology_blocks(bl, tec):
            if tec in technologies:
                bl.transfer_attributes_from(b_node.tech_blocks_new[tec])
            else:
                bl.transfer_attributes_from(b_node.tech_blocks_existing[tec])

        b_node.tech_blocks_active = pyo.Block(
            b_node.set_technologies, rule=init_active_technology_blocks
        )

        # Delete all auxiliary blocks
        if b_node.find_component("tech_blocks_new"):
            b_node.del_component(b_node.tech_blocks_new)
        if b_node.find_component("tech_blocks_new_index"):
            b_node.del_component(b_node.tech_blocks_new_index)
        if b_node.find_component("tech_blocks_existing"):
            b_node.del_component(b_node.tech_blocks_existing)
        if b_node.find_component("tech_blocks_existing_index"):
            b_node.del_component(b_node.tech_blocks_existing_index)

    def _define_solver_settings(self):
        """
        Defines solver and its settings depending on objective and solver
        """
        config = self.data.model_config
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        objective = config["optimization"]["objective"]["value"]

        # Set solver
        if config["solveroptions"]["solver"]["value"] in [
            "gurobi",
            "gurobi_persistent",
        ]:
            # Gurobi
            if not config["scaling"]["scaling_on"]["value"]:
                if objective in ["emissions_minC", "pareto"]:
                    config["solveroptions"]["solver"]["value"] = "gurobi_persistent"
            self.solver = get_gurobi_parameters(config["solveroptions"])

        elif config["solveroptions"]["solver"]["value"] == "glpk":
            self.solver = get_glpk_parameters(config["solveroptions"])

        # For persistent solver, set model instance
        if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
            self.solver.set_instance(model)

    def _optimize(self, objective):
        """
        Solves the model with the given objective
        """
        config = self.data.model_config

        # Define Objective Function
        if objective == "costs":
            self._optimize_cost()
        elif objective == "emissions_net":
            self._optimize_emissions_net()
        elif objective == "emissions_minC":
            self._optimize_costs_minE()
        elif objective == "costs_emissionlimit":
            self._optimize_costs_emissionslimit()
        else:
            raise Exception("objective in Configurations is incorrect")

        # Second stage of time averaging algorithm
        if config["optimization"]["timestaging"]["value"] != 0:
            self.info_solving_algorithms["time_stage"] = 2
            config["optimization"]["timestaging"]["value"] = 0
            self.info_solving_algorithms["objective"] = objective
            self._optimize_time_averaging_second_stage()

    def _optimize_cost(self):
        """
        Minimizes Costs
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        self._delete_objective()

        def init_cost_objective(obj):
            return model.var_npv

        model.objective = pyo.Objective(rule=init_cost_objective, sense=pyo.minimize)
        log_msg = "Set objective on cost"
        log.info(log_msg)
        self._call_solver()

    def _optimize_emissions_net(self):
        """
        Minimize net emissions
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        self._delete_objective()

        def init_emission_net_objective(obj):
            return model.var_emissions_net

        model.objective = pyo.Objective(
            rule=init_emission_net_objective, sense=pyo.minimize
        )
        log_msg = "Set objective on net emissions"
        log.info(log_msg)
        self._call_solver()

    def _optimize_costs_emissionslimit(self):
        """
        Minimize costs at emission limit
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        config = self.data.model_config

        emission_limit = config["optimization"]["emission_limit"]["value"]
        if model.find_component("const_emission_limit"):
            if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
                self.solver.remove_constraint(model.const_emission_limit)
            model.del_component(model.const_emission_limit)
        model.const_emission_limit = pyo.Constraint(
            expr=model.var_emissions_net <= emission_limit
        )
        if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
            self.solver.add_constraint(model.const_emission_limit)
        log_msg = "Defined constraint on net emissions"
        log.info(log_msg)
        self._optimize_cost()

    def _optimize_costs_minE(self):
        """
        Minimize costs at minimum emissions
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        config = self.data.model_config

        self._optimize_emissions_net()
        emission_limit = model.var_emissions_net.value
        if model.find_component("const_emission_limit"):
            if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
                self.solver.remove_constraint(model.const_emission_limit)
            model.del_component(model.const_emission_limit)
        model.const_emission_limit = pyo.Constraint(
            expr=model.var_emissions_net <= emission_limit * 1.001
        )
        if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
            self.solver.add_constraint(model.const_emission_limit)
        self._optimize_cost()

    def scale_model(self):
        """
        Creates a scaled model using the scale factors specified in the json files
        for technologies and networks as well as the global scaling factors
        specified. See also the documentation on model scaling.
        """
        config = self.data.model_config

        f_global = config["scaling"]["scaling_factors"]
        model_full = self.model[self.info_solving_algorithms["aggregation_model"]]

        model_full.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        # Scale technologies
        for period in model_full.periods:
            b_period = model_full.periods[period]
            # Scale technologies
            for node in b_period.node_blocks:
                for tec in b_period.node_blocks[node].tech_blocks_active:
                    b_tec = b_period.node_blocks[node].tech_blocks_active[tec]
                    model_full = self.data.technology_data[period][node][
                        tec
                    ].scale_model(b_tec, model_full, config)

            # Scale networks
            for netw in b_period.network_block:
                b_netw = b_period.network_block[netw]
                model_full = self.data.network_data[period][netw].scale_model(
                    b_netw, model_full, config
                )

            # Scale period
            if f_global["energy_vars"]["value"] >= 0:
                # Network constraints
                model_full.scaling_factor[
                    model_full.block_network_constraints[period].const_netw_inflow
                ] = f_global["energy_vars"]["value"]
                model_full.scaling_factor[
                    model_full.block_network_constraints[period].const_netw_outflow
                ] = f_global["energy_vars"]["value"]
                model_full.scaling_factor[
                    model_full.block_network_constraints[period].const_netw_consumption
                ] = f_global["energy_vars"]["value"]

                # Energy balance
                model_full.scaling_factor[
                    model_full.block_energybalance[period].const_energybalance
                ] = f_global["energy_vars"]["value"]

            # Costs balance
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_capex_tecs
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_capex_netw
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_opex_tecs
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_opex_netw
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost_tecs
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost_netws
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost_import
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost_export
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_violation_cost
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_revenue_carbon
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost_carbon
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])
            model_full.scaling_factor[
                model_full.block_costbalance[period].const_cost
            ] = (f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"])

            # Period Variables
            model_full.scaling_factor[b_period.var_cost_capex_tecs] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_capex_netws] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_opex_tecs] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_opex_netws] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_tecs] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_netws] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_imports] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_exports] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_violation] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_carbon_revenue] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_carbon_cost] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )
            model_full.scaling_factor[b_period.var_cost_total] = (
                f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
            )

            for node in b_period.node_blocks:
                b_node = b_period.node_blocks[node]
                model_full.scaling_factor[b_node.var_import_flow] = f_global[
                    "energy_vars"
                ]["value"]
                model_full.scaling_factor[b_node.var_export_flow] = f_global[
                    "energy_vars"
                ]["value"]
                model_full.scaling_factor[b_node.var_netw_inflow] = f_global[
                    "energy_vars"
                ]["value"]
                model_full.scaling_factor[b_node.var_netw_outflow] = f_global[
                    "energy_vars"
                ]["value"]
                if b_node.find_component("var_netw_consumption"):
                    model_full.scaling_factor[b_node.var_netw_consumption] = f_global[
                        "energy_vars"
                    ]["value"]
                model_full.scaling_factor[b_node.var_generic_production] = f_global[
                    "energy_vars"
                ]["value"]
                model_full.scaling_factor[b_node.const_generic_production] = f_global[
                    "energy_vars"
                ]["value"]
                model_full.scaling_factor[b_node.var_import_flow] = f_global[
                    "energy_vars"
                ]["value"]

        # Global cost balance
        model_full.scaling_factor[model_full.const_npv] = (
            f_global["cost_vars"]["value"] * f_global["energy_vars"]["value"]
        )
        # Scale objective
        model_full.scaling_factor[model_full.objective] = (
            f_global["objective"]["value"] * f_global["cost_vars"]["value"]
        )

        self.model["scaled"] = pyo.TransformationFactory(
            "core.scale_model"
        ).create_using(model_full)

    def _call_solver(self):
        """
        Calls the solver and solves the model
        """
        log.info("Solving Model...")

        start = time.time()
        config = self.data.model_config

        # Create save path and folder
        time_stamp = datetime.datetime.fromtimestamp(start).strftime("%Y%m%d%H%M%S")
        save_path = Path(config["reporting"]["save_path"]["value"])

        if config["reporting"]["case_name"]["value"] == -1:
            folder_name = str(time_stamp)
        else:
            folder_name = (
                str(time_stamp) + "_" + config["reporting"]["case_name"]["value"]
            )
        if self.info_pareto["pareto_point"]:
            folder_name = folder_name + str(self.info_pareto["pareto_point"])

        result_folder_path = create_unique_folder_name(save_path, folder_name)
        create_save_folder(result_folder_path)

        # Scale model
        if config["scaling"]["scaling_on"]["value"] == 1:
            self.scale_model()
            model = self.model["scaled"]
        else:
            model = self.model[self.info_solving_algorithms["aggregation_model"]]

        # Call solver
        if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
            self.solver.set_objective(model.objective)

        if config["solveroptions"]["solver"]["value"] == "glpk":
            self.solution = self.solver.solve(
                model,
                tee=True,
                logfile=str(Path(result_folder_path / "solver_log.txt")),
                keepfiles=True,
            )
        else:
            self.solution = self.solver.solve(
                model,
                tee=True,
                warmstart=True,
                logfile=str(Path(result_folder_path / "solver_log.txt")),
                keepfiles=True,
            )

        # Determine if results should be written
        if "write_results" in config["reporting"].keys():
            if config["reporting"]["write_results"]["value"] == 1:
                write_results = True
            else:
                write_results = False
        else:
            warnings.warn(
                "The config file needs to contain config['reporting']"
                "['write_results']. This is mandatory in future versions",
                FutureWarning,
                stacklevel=2,
            )
            write_results = True

        # Check if solution is available
        if write_results:
            if (self.solution.solver.status == pyo.SolverStatus.ok) or (
                self.solution.solver.status == pyo.SolverStatus.warning
            ):
                write_results = True
            if self.solution.solver.termination_condition in [
                pyo.TerminationCondition.infeasibleOrUnbounded,
                pyo.TerminationCondition.infeasible,
                pyo.TerminationCondition.unbounded,
            ]:
                write_results = False

        if write_results:
            if config["scaling"]["scaling_on"]["value"] == 1:
                pyo.TransformationFactory("core.scale_model").propagate_solution(
                    model, self.model[self.info_solving_algorithms["aggregation_model"]]
                )

        if config["reporting"]["write_solution_diagnostics"]["value"] >= 1:
            self._write_solution_diagnostics(result_folder_path)

        self.solution.write()

        self.last_solve_info["pareto_point"] = self.info_pareto["pareto_point"]
        self.last_solve_info["monte_carlo_run"] = self.info_monte_carlo[
            "monte_carlo_run"
        ]
        self.last_solve_info["config"] = config
        self.last_solve_info["result_folder_path"] = result_folder_path
        self.last_solve_info["time_stage"] = self.info_solving_algorithms["time_stage"]

        # Write results to path
        if write_results:
            self.write_results()

        log.info("Solving model completed in " + str(round(time.time() - start)) + " s")

    def _write_solution_diagnostics(self, save_path):
        """
        Can write solution quality, constraint map and variable map to file. Options
        are specified in the configuration.

        :param save_path:
        :return:
        """
        config = self.data.model_config
        model = self.solver._solver_model
        constraint_map = self.solver._pyomo_con_to_solver_con_map
        variable_map = self.solver._pyomo_var_to_solver_var_map

        # Write solution quality to txt
        with open(f"{save_path}/diag_solution_quality.txt", "w") as file:
            sys.stdout = file  # Redirect stdout to the file
            model.printQuality()  # Call the function that prints something
            sys.stdout = sys.__stdout__  # Reset stdout to the console

        if config["reporting"]["write_solution_diagnostics"]["value"] >= 2:
            # Write constraint map to txt
            with open(f"{save_path}/diag_constraint_map.txt", "w") as file:
                for key, value in constraint_map.items():
                    file.write(f"{key}: {value}\n")

            # Write var map to txt
            with open(f"{save_path}/diag_variable_map.txt", "w") as file:
                for key, value in variable_map._dict.items():
                    file.write(f"{value[0].name}: {value[1]}\n")

    def _solve_pareto(self):
        """
        Optimize the pareto front
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]
        config = self.data.model_config
        pareto_points = config["optimization"]["pareto_points"]["value"]

        # Min Emissions
        self.info_pareto["pareto_point"] = pareto_points
        self._optimize_costs_minE()
        emissions_min = model.var_emissions_net.value

        # Min Cost
        self.info_pareto["pareto_point"] = 1
        self._optimize_cost()
        emissions_max = model.var_emissions_net.value

        # Emission limit
        emission_limits = np.linspace(emissions_max, emissions_min, num=pareto_points)[
            1:-1
        ]

        for limit in range(0, len(emission_limits)):
            self.info_pareto["pareto_point"] += 1
            log_msg = f"Optimizing Pareto point {limit}"
            log.info(log_msg)
            if limit != 0:
                # If its not the first point, delete constraint
                if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
                    self.solver.remove_constraint(model.const_emission_limit)
                model.del_component(model.const_emission_limit)
            model.const_emission_limit = pyo.Constraint(
                expr=model.var_emissions_net <= emission_limits[limit] * 1.005
            )
            if config["solveroptions"]["solver"]["value"] == "gurobi_persistent":
                self.solver.add_constraint(model.const_emission_limit)
            self._optimize("costs")

    def _solve_monte_carlo(self, objective: str):
        """
        Optimizes multiple runs with monte carlo

        :param str objective: objective to optimize
        """
        config = self.data.model_config
        self.info_monte_carlo["monte_carlo_run"] = 0

        for run in range(0, config["optimization"]["monte_carlo"]["N"]["value"]):
            self.info_monte_carlo["monte_carlo_run"] = run
            self._monte_carlo_set_cost_parameters()
            if run == 0:
                # in this case we need to set the objective
                self._optimize(objective)
            else:
                # in this case we can call the solver directly
                self._call_solver()

        summary_path = Path.joinpath(
            Path(config["reporting"]["save_summary_path"]["value"]), "Summary.xlsx"
        )
        if config["optimization"]["monte_carlo"]["type"]["value"] == "normal_dis":
            component_set = config["optimization"]["monte_carlo"]["on_what"]["value"]
        elif (
            config["optimization"]["monte_carlo"]["type"]["value"]
            == "uniform_dis_from_file"
        ):
            component_set = list(set(self.data.monte_carlo_specs["type"]))
        add_values_to_summary(summary_path, component_set=component_set)

    def _monte_carlo_set_cost_parameters(self):
        """
        Changes cost parameters for monte carlo analysis.
        """
        config = self.data.model_config

        # use correct resolution
        model = self.model[self.info_solving_algorithms["aggregation_model"]]
        monte_carlo_type = config["optimization"]["monte_carlo"]["type"]["value"]
        monte_carlo_on = config["optimization"]["monte_carlo"]["on_what"]["value"]

        import_constraint_reconstruction = False
        export_constraint_reconstruction = False

        if monte_carlo_type == "normal_dis":
            if "Technologies" in monte_carlo_on:
                for period in model.periods:
                    for node in model.periods[period].node_blocks:
                        for tec in (
                            model.periods[period].node_blocks[node].tech_blocks_active
                        ):
                            self._monte_carlo_technologies(period, node, tec)

            if "Networks" in monte_carlo_on:
                for period in model.periods:
                    for netw in model.periods[period].network_block:
                        self._monte_carlo_networks(period, netw)

            if "Import" in monte_carlo_on:
                self._monte_carlo_import_parameters()
                import_constraint_reconstruction = True

            if "Export" in monte_carlo_on:
                self._monte_carlo_export_parameters()
                export_constraint_reconstruction = True

        elif monte_carlo_type == "uniform_dis_from_file":
            MC_parameters = self.data.monte_carlo_specs
            processed_names = set()

            for index, row in MC_parameters.iterrows():
                if row["type"] == "Technologies":
                    tec = row["name"]

                    # Iterate through periods and nodes in the model
                    if tec not in processed_names:
                        for period in model.periods:
                            for node in model.periods[period].node_blocks:
                                tech_blocks = (
                                    model.periods[period]
                                    .node_blocks[node]
                                    .tech_blocks_active
                                )

                                # Check if the technology is active in the current node
                                if tec in tech_blocks:
                                    capex_model = self.data.technology_data[period][
                                        node
                                    ][tec].economics.capex_model

                                    if capex_model == 1:
                                        if row["parameter"] == "unit_CAPEX":
                                            self._monte_carlo_technologies(
                                                period, node, tec, row
                                            )
                                        else:
                                            new_row = MC_parameters[
                                                (MC_parameters["name"] == tec)
                                                & (
                                                    MC_parameters["parameter"]
                                                    == "unit_CAPEX"
                                                )
                                            ]
                                            if not new_row.empty:
                                                self._monte_carlo_technologies(
                                                    period, node, tec, new_row
                                                )
                                            else:
                                                log_msg = f"Parameter unit_CAPEX is not defined for {tec} in MonteCarlo.csv"
                                                log.warning(log_msg)

                                    else:
                                        # Find all rows with the same technology name
                                        MC_technology_rows = MC_parameters[
                                            (MC_parameters["type"] == "Technologies")
                                            & (MC_parameters["name"] == tec)
                                        ]
                                        self._monte_carlo_technologies(
                                            period, node, tec, MC_technology_rows
                                        )

                                    processed_names.add(tec)
                                    break
                                else:
                                    log_msg = f"Technology {tec} in MonteCarlo.csv is not an active component"
                                    log.warning(log_msg)

                elif row["type"] == "Networks":
                    netw = row["name"]
                    if netw not in processed_names:
                        for period in model.periods:
                            netw_blocks = model.periods[period].network_block
                            if netw in netw_blocks:
                                MC_network_rows = MC_parameters[
                                    (MC_parameters["type"] == "Networks")
                                    & (MC_parameters["name"] == netw)
                                ]

                                self._monte_carlo_networks(
                                    period, netw, MC_network_rows
                                )

                                processed_names.add(netw)
                                break
                            else:
                                log_msg = f"Network {netw} in MonteCarlo.csv is not active component"
                                log.warning(log_msg)

                elif row["type"] == "Import":
                    import_constraint_reconstruction = True
                    car = row["name"]
                    self._monte_carlo_import_parameters(car, row)
                elif row["type"] == "Export":
                    export_constraint_reconstruction = True
                    car = row["name"]
                    self._monte_carlo_export_parameters(car, row)

        if import_constraint_reconstruction:
            self._monte_carlo_import_constraints()

        if export_constraint_reconstruction:
            self._monte_carlo_export_constraints()

    def _monte_carlo_technologies(self, period, node, tec, MC_ranges=None):
        """
        Changes the capex of technologies
        """
        aggregation_model = self.info_solving_algorithms["aggregation_model"]
        aggregation_data = self.info_solving_algorithms["aggregation_data"]

        config = self.data.model_config
        tec_data = self.data.technology_data[period][node][tec]
        model = self.model[aggregation_model]

        if tec_data.economics.capex_model in [1, 3]:
            # Preprocessing
            sd = config["optimization"]["monte_carlo"]["sd"]["value"]
            sd_random = np.random.normal(1, sd)

            economics = tec_data.economics
            discount_rate = set_discount_rate(config, economics)
            fraction_of_year_modelled = self.data.topology["fraction_of_year_modelled"]

            b_tec = model.periods[period].node_blocks[node].tech_blocks_active[tec]

            annualization_factor = annualize(
                discount_rate, economics.lifetime, fraction_of_year_modelled
            )

            # Change parameters
            if tec_data.economics.capex_model == 1:
                # UNIT CAPEX
                # Update parameter
                if MC_ranges is not None:
                    if isinstance(MC_ranges, pd.Series):
                        unit_capex = random.uniform(MC_ranges["min"], MC_ranges["max"])
                    elif isinstance(MC_ranges, pd.DataFrame):
                        unit_capex = random.uniform(
                            MC_ranges["min"].iloc[0], MC_ranges["max"].iloc[0]
                        )
                else:
                    unit_capex = tec_data.economics.capex_data["unit_capex"] * sd_random

                b_tec.para_unit_capex = unit_capex
                b_tec.para_unit_capex_annual = unit_capex * annualization_factor

            elif tec_data.economics.capex_model == 3:
                if MC_ranges is not None:
                    for _, row in MC_ranges.iterrows():
                        if row["parameter"] == "unit_CAPEX":
                            unit_capex = random.uniform(row["min"], row["max"])
                        elif row["parameter"] == "fix_CAPEX":
                            fix_capex = random.uniform(row["min"], row["max"])
                else:
                    unit_capex = tec_data.economics.capex_data["unit_capex"] * sd_random
                    fix_capex = tec_data.economics.capex_data["fix_capex"] * sd_random

                b_tec.para_unit_capex = unit_capex
                b_tec.para_unit_capex_annual = unit_capex * annualization_factor
                b_tec.para_fix_capex = fix_capex
                b_tec.para_fix_capex_annual = fix_capex * annualization_factor

            # Change variable bounds
            def calculate_max_capex():
                if economics.capex_model == 1:
                    max_capex = b_tec.para_size_max * b_tec.para_unit_capex_annual
                    bounds = (0, max_capex)
                elif economics.capex_model == 3:
                    max_capex = (
                        b_tec.para_size_max * b_tec.para_unit_capex_annual
                        + b_tec.para_fix_capex_annual
                    )
                    bounds = (0, max_capex)
                else:
                    bounds = None
                return bounds

            bounds = calculate_max_capex()
            b_tec.var_capex_aux.setlb(bounds[0])
            b_tec.var_capex_aux.setub(bounds[1])

            # Delete constraints/conjunctions/relaxations
            if economics.capex_model == 1:
                big_m_transformation_required = 0
                b_tec.del_component(b_tec.const_capex_aux)
                b_tec.del_component(b_tec.const_capex)
            elif economics.capex_model == 3:
                big_m_transformation_required = 1
                b_tec.del_component(b_tec.dis_installation)
                b_tec.del_component(b_tec.disjunction_installation)
                b_tec.del_component(b_tec._pyomo_gdp_bigm_reformulation)

            # Reconstruct technology constraints
            data_period = get_data_for_investment_period(
                self.data, period, aggregation_data
            )
            data_node = get_data_for_node(data_period, node)

            b_tec = tec_data._define_capex_constraints(b_tec, data_node)
            if big_m_transformation_required:
                b_tec = perform_disjunct_relaxation(b_tec)

        else:
            log_msg = (
                "monte carlo for capex models other than 1 and 3 is not implemented"
            )
            log.warning(log_msg)

    def _monte_carlo_networks(self, period, netw, MC_ranges=None):
        """
        Changes the capex of networks
        """
        aggregation_model = self.info_solving_algorithms["aggregation_model"]

        config = self.data.model_config
        model = self.model[aggregation_model]

        sd = config["optimization"]["monte_carlo"]["sd"]["value"]
        sd_random = np.random.normal(1, sd)

        netw_data = self.data.network_data[period][netw]
        economics = netw_data.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = self.data.topology["fraction_of_year_modelled"]

        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        b_netw = model.periods[period].network_block[netw]

        if MC_ranges is not None:
            for _, row in MC_ranges.iterrows():
                if row["parameter"] == "gamma1":
                    b_netw.para_capex_gamma1 = random.uniform(row["min"], row["max"])
                elif row["parameter"] == "gamma2":
                    b_netw.para_capex_gamma2 = random.uniform(row["min"], row["max"])
                elif row["parameter"] == "gamma3":
                    b_netw.para_capex_gamma3 = random.uniform(row["min"], row["max"])
                elif row["parameter"] == "gamma4":
                    b_netw.para_capex_gamma4 = random.uniform(row["min"], row["max"])
        else:
            # Update cost parameters
            b_netw.para_capex_gamma1 = (
                economics.capex_data["gamma1"] * annualization_factor * sd_random
            )
            b_netw.para_capex_gamma2 = (
                economics.capex_data["gamma2"] * annualization_factor * sd_random
            )
            b_netw.para_capex_gamma3 = (
                economics.capex_data["gamma3"] * annualization_factor * sd_random
            )
            b_netw.para_capex_gamma4 = (
                economics.capex_data["gamma4"] * annualization_factor * sd_random
            )

        for arc in b_netw.set_arcs:
            b_arc = b_netw.arc_block[arc]

            def calculate_max_capex():
                max_capex = (
                    b_netw.para_capex_gamma1
                    + b_netw.para_capex_gamma2 * b_arc.para_size_max
                    + b_netw.para_capex_gamma3 * b_arc.distance
                    + b_netw.para_capex_gamma4 * b_arc.para_size_max * b_arc.distance
                )
                return (0, max_capex)

            bounds = calculate_max_capex()
            b_arc.var_capex_aux.setlb(bounds[0])
            b_arc.var_capex_aux.setub(bounds[1])
            b_arc.var_capex.setlb(bounds[0])
            b_arc.var_capex.setub(bounds[1])

            # Remove constraint (from persistent solver and from model)
            if b_arc.find_component("dis_installation"):
                b_arc.del_component(b_arc._pyomo_gdp_bigm_reformulation)
                b_arc.del_component(b_arc.dis_installation)
                b_arc.del_component(b_arc.disjunction_installation)
            if b_arc.find_component("const_capex_aux"):
                b_arc.del_component(b_arc.const_capex_aux)
            b_arc.del_component(b_arc.const_capex)

            b_arc = netw_data._define_capex_constraints_arc(
                b_arc, b_netw, arc[0], arc[1]
            )

            if b_arc.big_m_transformation_required:
                b_arc = perform_disjunct_relaxation(b_arc)

    def _monte_carlo_import_parameters(self, on_car=None, MC_ranges=None):
        """
        Changes the import prices
        """
        aggregation_model = self.info_solving_algorithms["aggregation_model"]
        aggregation_data = self.info_solving_algorithms["aggregation_data"]

        config = self.data.model_config
        model = self.model[aggregation_model]

        for period in model.periods:
            b_period = model.periods[period]
            set_t = get_set_t(config, b_period)

            for node in b_period.node_blocks:
                if on_car is None:
                    # change for all carriers at node
                    on_car = model.periods[period].node_blocks[node].set_carriers

                for car in model.periods[period].node_blocks[node].set_carriers:
                    if car in on_car:
                        import_prices = self.data.time_series[aggregation_data][
                            period, node, "CarrierData", car, "Import price"
                        ]

                        for t in set_t:
                            if MC_ranges is None:
                                sd = config["optimization"]["monte_carlo"]["sd"][
                                    "value"
                                ]
                                random_factor = np.random.normal(1, sd)
                            else:
                                random_factor = random.uniform(
                                    MC_ranges["min"], MC_ranges["max"]
                                )

                            # Update parameter
                            model.periods[period].node_blocks[node].para_import_price[
                                t, car
                            ] = (import_prices.iloc[t - 1] * random_factor)

    def _monte_carlo_import_constraints(self):
        """
        Reconstructs the import cost constraint
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        # reconstruct constraint
        for period in model.periods:
            b_period = model.periods[period]
            b_period_cost = model.block_costbalance[period]

            b_period_cost.del_component(b_period_cost.const_cost_import)
            b_period_cost.const_cost_import = construct_import_costs(
                b_period, self.data, period
            )

    def _monte_carlo_export_parameters(self, on_car=None, MC_ranges=None):
        """
        Changes the import prices
        """
        aggregation_model = self.info_solving_algorithms["aggregation_model"]
        aggregation_data = self.info_solving_algorithms["aggregation_data"]

        config = self.data.model_config
        model = self.model[aggregation_model]

        for period in model.periods:
            b_period = model.periods[period]
            set_t = get_set_t(config, b_period)

            for node in b_period.node_blocks:
                if on_car is None:
                    # change for all carriers at node
                    on_car = model.periods[period].node_blocks[node].set_carriers

                for car in model.periods[period].node_blocks[node].set_carriers:
                    if car in on_car:
                        export_prices = self.data.time_series[aggregation_data][
                            period, node, "CarrierData", car, "Export price"
                        ]

                        for t in set_t:
                            if MC_ranges is None:
                                sd = config["optimization"]["monte_carlo"]["sd"][
                                    "value"
                                ]
                                random_factor = np.random.normal(1, sd)
                            else:
                                random_factor = random.uniform(
                                    MC_ranges["min"], MC_ranges["max"]
                                )

                            # Update parameter
                            model.periods[period].node_blocks[node].para_export_price[
                                t, car
                            ] = (export_prices.iloc[t - 1] * random_factor)

    def _monte_carlo_export_constraints(self):
        """
        Reconstructs the import cost constraint
        """
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        # reconstruct constraint
        for period in model.periods:
            b_period = model.periods[period]
            b_period_cost = model.block_costbalance[period]

            b_period_cost.del_component(b_period_cost.const_cost_export)
            b_period_cost.const_cost_export = construct_export_costs(
                b_period, self.data, period
            )

    def _delete_objective(self):
        """
        Delete the objective function
        """
        config = self.data.model_config
        model = self.model[self.info_solving_algorithms["aggregation_model"]]

        if not config["optimization"]["monte_carlo"]["N"]["value"]:
            try:
                model.del_component(model.objective)
            except:
                pass

    def _optimize_time_averaging_second_stage(self):
        """
        Optimizes the second stage of the time_averaging algorithm
        """
        # Todo: make it possible to chose (config)
        bounds_on = "no_storage"
        self.construct_model()
        self.construct_balances()
        self._impose_size_constraints(bounds_on)
        self._optimize(self.info_solving_algorithms["objective"])

    def _impose_size_constraints(self, bounds_on):
        """
        Formulates lower bound on technology and network sizes.

        It is possible to exclude storage technologies or networks by specifying
        bounds_on.

        :param bounds_on: can be 'all', 'only_technologies', 'only_networks', 'no_storage'
        """

        m_full = self.model["full"]
        m_avg = self.model["averaged"]

        # Technologies
        if (
            bounds_on == "all"
            or bounds_on == "only_technologies"
            or bounds_on == "no_storage"
        ):

            def size_constraint_block_tecs_init(block, period, node):
                def size_constraints_tecs_init(const, tec):
                    if (
                        self.data.technology_data[period][node][
                            tec
                        ].component_options.technology_model
                        == "STOR"
                        and bounds_on == "no_storage"
                    ):
                        return pyo.Constraint.Skip
                    elif self.data.technology_data[period][node][tec].existing:
                        return pyo.Constraint.Skip
                    else:
                        log_msg = (
                            f"Size constraint imposed on {tec} at {node} in {period}"
                        )
                        log.info(log_msg)

                        return (
                            m_avg.periods[period]
                            .node_blocks[node]
                            .tech_blocks_active[tec]
                            .var_size.value
                            <= m_full.periods[period]
                            .node_blocks[node]
                            .tech_blocks_active[tec]
                            .var_size
                        )

                block.size_constraints_tecs = pyo.Constraint(
                    m_full.periods[period].node_blocks[node].set_technologies,
                    rule=size_constraints_tecs_init,
                )

            m_full.size_constraint_tecs = pyo.Block(
                m_full.set_periods,
                m_full.set_nodes,
                rule=size_constraint_block_tecs_init,
            )

        # Networks
        if (
            bounds_on == "all"
            or bounds_on == "only_networks"
            or bounds_on == "no_storage"
        ):

            def size_constraint_block_netw_init(block, period):

                def size_constraints_netw_init(const, netw):

                    b_netw_full = m_full.periods[period].network_block[netw]
                    b_netw_avg = m_avg.periods[period].network_block[netw]

                    log_msg = f"Size constraint imposed on {netw} in {period}"
                    log.info(log_msg)

                    def size_constraints_arcs_init(const, node_from, node_to):
                        return (
                            b_netw_full.arc_block[node_from, node_to].var_size
                            >= b_netw_avg.arc_block[node_from, node_to].var_size.value
                        )

                    block.size_constraints_arcs = pyo.Constraint(
                        b_netw_full.set_arcs, rule=size_constraints_arcs_init
                    )

                block.size_constraints_netw = pyo.Block(
                    m_full.periods[period].set_networks, rule=size_constraints_netw_init
                )

            m_full.size_constraints_netw = pyo.Block(
                m_full.set_periods, rule=size_constraint_block_netw_init
            )



import pyomo.environ as pyo
import pyomo.gdp as gdp
from warnings import warn
import pandas as pd


from ..genericTechnologies.fitting_classes import (
    FitGenericTecTypeType1,
    FitGenericTecTypeType2,
    FitGenericTecTypeType34,
)
from ..technology import Technology
from ...utilities import link_full_resolution_to_clustered


class Conv3(Technology):
    """
    Technology with no input/output substitution

    This technology type resembles a technology for which the output can be written
    as a function of the input, according to different performance functions that
    can be specified in the JSON files (``performance_function_type``).
    Four different performance function fits of the technology data (again specified
    in the JSON file) are possible, and for all the function is based on the input
    of the main carrier , i.e.,: :math:`output_{car} = f_{car}(input_{maincarrier})`.
    Note that the ratio between all input carriers is fixed.

    **Constraint declarations:**

    For all technologies modelled with CONV3 (regardless of performance function type):
    - Size constraints are formulated on the input.

      .. math::
         Input_{t, maincarrier} \\leq S

    - The ratios of inputs are fixed and given as:

      .. math::
        Input_{t, car} = {\\phi}_{car} * Input_{t, maincarrier}

      If the technology is turned off, all inputs are set to zero.

    - ``performance_function_type == 1``: Linear through origin. Note that if
      min_part_load is larger 0, the technology cannot be turned off.

      .. math::
        Output_{t, car} = {\\alpha}_{1, car} Input_{t, maincarrier}

      .. math::
        min_part_load * S \\leq {\\alpha}_1 Input_{t, maincarrier}

    - ``performance_function_type == 2``: Linear with minimal partload. If the
      technology is in on, it holds:

      If the technology is in on, it holds:

      .. math::
        Output_{t, car} = {\\alpha}_{1, car} Input_{t, maincarrier} + {\\alpha}_{2, car}

      .. math::
        Input_{maincarrier} \\geq Input_{min} * S

    - If the technology is off, input and output are set to 0:

      .. math::
         Output_{t, car} = 0

      .. math::
         Input_{t, maincarrier} = 0

    - ``performance_function_type == 3``: Piecewise linear performance function (
      makes big-m transformation required). The same constraints as for
      ``performance_function_type == 2`` with the exception that the performance
      function is defined piecewise for the respective number of pieces.

    - ``performance_function_type == 4``:Piece-wise linear, minimal partload. Enables the modeling
      of technologies with slow (>1h) startup and shutdown trajectories. For more information
      please refer to dynamics under advanced topics. Based on Equations 9-11, 13 and 15 in Morales-España, G., Ramírez-Elizondo, L.,
      & Hobbs, B. F. (2017). Hidden power system inflexibilities imposed by
      traditional unit commitment formulations. Applied Energy, 191, 223–238.
      https://doi.org/10.1016/J.APENERGY.2017.01.089

    - Additionally, ramping rates of the technology can be constrained.

      .. math::
         -rampingrate \\leq Input_{t, main-car} - Input_{t-1, car} \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.size_based_on = "input"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

        # Initialize fitting class
        if self.component_options.performance_function_type == 1:
            self.fitting_class = FitGenericTecTypeType1(self.component_options)
        elif self.component_options.performance_function_type == 2:
            self.fitting_class = FitGenericTecTypeType2(self.component_options)
        elif (
            self.component_options.performance_function_type == 3
            or self.component_options.performance_function_type == 4
        ):
            self.fitting_class = FitGenericTecTypeType34(self.component_options)
        else:
            raise Exception(
                "performance_function_type must be an integer between 1 and 4"
            )

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits conversion technology type 3

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(Conv3, self).fit_technology_performance(climate_data, location)

        if self.component_options.size_based_on == "output":
            raise Exception("size_based_on == output for CONV3 not possible.")

        # fit coefficients
        self.processed_coeff.time_independent["fit"] = (
            self.fitting_class.fit_performance_function(
                self.input_parameters.performance_data["performance"]
            )
        )

        phi = {}
        for car in self.input_parameters.performance_data["input_ratios"]:
            phi[car] = self.input_parameters.performance_data["input_ratios"][car]
        self.processed_coeff.time_independent["phi"] = phi

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(Conv3, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        self.bounds["input"] = self.fitting_class.calculate_input_bounds(
            self.component_options.size_based_on, time_steps
        )
        self.bounds["output"] = self.fitting_class.calculate_output_bounds(
            self.component_options.size_based_on, time_steps
        )

        # Input bounds recalculation
        for car in self.component_options.input_carrier:
            if not car == self.component_options.main_input_carrier:
                self.bounds["input"][car] = (
                    self.bounds["input"][self.component_options.main_input_carrier]
                    * self.input_parameters.performance_data["input_ratios"][car]
                )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type CONV3

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(Conv3, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        # DATA OF TECHNOLOGY
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics
        rated_power = self.input_parameters.rated_power

        if self.component_options.performance_function_type == 1:
            b_tec = self._performance_function_type_1(b_tec)
        elif self.component_options.performance_function_type == 2:
            b_tec = self._performance_function_type_2(b_tec)
        elif self.component_options.performance_function_type == 3:
            b_tec = self._performance_function_type_3(b_tec)
        elif self.component_options.performance_function_type == 4:
            b_tec = self._performance_function_type_4(b_tec)

        # Size constraints
        # constraint on input ratios
        standby_power = coeff_ti["standby_power"]
        phi = coeff_ti["phi"]

        if self.component_options.performance_function_type == 1 or standby_power == -1:

            def init_input_input(const, t, car_input):
                if car_input == self.component_options.main_input_carrier:
                    return pyo.Constraint.Skip
                else:
                    return (
                        self.input[t, car_input]
                        == phi[car_input]
                        * self.input[t, self.component_options.main_input_carrier]
                    )

            b_tec.const_input_input = pyo.Constraint(
                self.set_t_performance, b_tec.set_input_carriers, rule=init_input_input
            )
        else:

            self.big_m_transformation_required = 1

            if self.component_options.standby_power_carrier == -1:
                car_standby_power = self.component_options.main_input_carrier
            else:
                car_standby_power = self.component_options.standby_power_carrier

            s_indicators = range(0, 2)

            def init_input_input(dis, t, ind):
                if ind == 0:  # technology off
                    dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                    def init_input_off(const, car_input):
                        if car_input == car_standby_power:
                            return pyo.Constraint.Skip
                        else:
                            return self.input[t, car_input] == 0

                    dis.const_input_off = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_input_off
                    )

                else:  # technology on
                    dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                    def init_input_on(const, car_input):
                        if car_input == self.component_options.main_input_carrier:
                            return pyo.Constraint.Skip
                        else:
                            return (
                                self.input[t, car_input]
                                == phi[car_input]
                                * self.input[
                                    t, self.component_options.main_input_carrier
                                ]
                            )

                    dis.const_input_on = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_input_on
                    )

            b_tec.dis_input_input = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_input_input
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_input_input[t, i] for i in s_indicators]

            b_tec.disjunction_input_input = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        # size constraint based on main carrier input
        def init_size_constraint(const, t):
            return (
                self.input[t, self.component_options.main_input_carrier]
                <= b_tec.var_size * rated_power
            )

        b_tec.const_size = pyo.Constraint(
            self.set_t_performance, rule=init_size_constraint
        )

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def _performance_function_type_1(self, b_tec):
        """
        Linear, through origin, min partload possible

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """

        # Performance parameters:
        rated_power = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent
        alpha1 = {}
        for car in coeff_ti["fit"]:
            alpha1[car] = coeff_ti["fit"][car]["alpha1"]
        min_part_load = coeff_ti["min_part_load"]

        # Input-output relation
        def init_input_output(const, t, car_output):
            return (
                self.output[t, car_output]
                == alpha1[car_output]
                * self.input[t, self.component_options.main_input_carrier]
            )

        b_tec.const_input_output = pyo.Constraint(
            self.set_t_performance, b_tec.set_output_carriers, rule=init_input_output
        )

        # setting the minimum part load constraint if applicable
        if min_part_load > 0:

            def init_min_part_load(const, t):
                return (
                    min_part_load * b_tec.var_size * rated_power
                    <= self.input[t, self.component_options.main_input_carrier]
                )

            b_tec.const_min_part_load = pyo.Constraint(
                self.set_t_performance, rule=init_min_part_load
            )

        return b_tec

    def _performance_function_type_2(self, b_tec):
        """
        Linear, minimal partload

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """

        # Transformation required
        self.big_m_transformation_required = 1

        # Performance Parameters
        rated_power = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent
        alpha1 = {}
        alpha2 = {}
        for car in coeff_ti["fit"]:
            alpha1[car] = coeff_ti["fit"][car]["alpha1"]
            alpha2[car] = coeff_ti["fit"][car]["alpha2"]
        min_part_load = coeff_ti["min_part_load"]
        standby_power = coeff_ti["standby_power"]

        # Performance Parameters

        if standby_power != -1:
            if self.component_options.standby_power_carrier == -1:
                car_standby_power = self.component_options.main_input_carrier
            else:
                car_standby_power = self.component_options.standby_power_carrier

        if not b_tec.find_component("var_x"):
            b_tec.var_x = pyo.Var(
                self.set_t_performance, domain=pyo.NonNegativeReals, bounds=(0, 1)
            )

        if min_part_load == 0:
            warn(
                "Having performance_function_type = 2 with no part-load usually makes no sense. Error occured for "
                + self.name
            )

        # define disjuncts
        s_indicators = range(0, 2)

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                if standby_power == -1:

                    def init_input_off(const, car_input):
                        return self.input[t, car_input] == 0

                    dis.const_input = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_input_off
                    )

                else:

                    def init_standby_power(const, car_input):
                        if car_input == self.component_options.main_input_carrier:
                            return (
                                self.input[t, car_standby_power]
                                == standby_power * b_tec.var_size * rated_power
                            )
                        else:
                            return self.input[t, car_input] == 0

                    dis.const_input = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_standby_power
                    )

                def init_output_off(const, car_output):
                    return self.output[t, car_output] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

            else:  # technology on

                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                # input-output relation
                def init_input_output_on(const, car_output):
                    return (
                        self.output[t, car_output]
                        == alpha1[car_output]
                        * self.input[t, self.component_options.main_input_carrier]
                        + alpha2[car_output] * b_tec.var_size * rated_power
                    )

                dis.const_input_output_on = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_input_output_on
                )

                # min part load constraint
                def init_min_partload(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec

    def _performance_function_type_3(self, b_tec):
        """
        Sets the input-output constraint for a tec based on tec_type CONV3 with performance type 3.

        Type 3 is a piecewise linear fit to the performance data, based on the number of segments specified. Note that
        this requires a big-m transformation. Again, a minimum part load is possible.

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        # Transformation required
        self.big_m_transformation_required = 1

        # Performance Parameters
        rated_power = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent
        alpha1 = {}
        alpha2 = {}
        for car in coeff_ti["fit"]:
            bp_x = coeff_ti["fit"][car]["bp_x"]
            alpha1[car] = coeff_ti["fit"][car]["alpha1"]
            alpha2[car] = coeff_ti["fit"][car]["alpha2"]
        min_part_load = coeff_ti["min_part_load"]
        standby_power = coeff_ti["standby_power"]

        if standby_power != -1:
            if self.component_options.standby_power_carrier == -1:
                car_standby_power = self.component_options.main_input_carrier
            else:
                car_standby_power = self.component_options.standby_power_carrier

        if not b_tec.find_component("var_x"):
            b_tec.var_x = pyo.Var(
                self.set_t_performance, domain=pyo.NonNegativeReals, bounds=(0, 1)
            )

        s_indicators = range(0, len(bp_x))

        def init_input_output(dis, t, ind):
            if ind == 0:  # technology off

                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                if standby_power == -1:

                    def init_input_off(const, car_input):
                        return self.input[t, car_input] == 0

                    dis.const_input_off = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_input_off
                    )

                else:

                    def init_standby_power(const, car_input):
                        if car_input == self.component_options.main_input_carrier:
                            return (
                                self.input[t, car_standby_power]
                                == standby_power * b_tec.var_size * rated_power
                            )

                        else:
                            return self.input[t, car_input] == 0

                    dis.const_input = pyo.Constraint(
                        b_tec.set_input_carriers, rule=init_standby_power
                    )

                def init_output_off(const, car_output):
                    return self.output[t, car_output] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

            else:  # piecewise definition

                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                def init_input_on1(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        >= bp_x[ind - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_on1 = pyo.Constraint(rule=init_input_on1)

                def init_input_on2(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        <= bp_x[ind] * b_tec.var_size * rated_power
                    )

                dis.const_input_on2 = pyo.Constraint(rule=init_input_on2)

                def init_output_on(const, car_output):
                    return (
                        self.output[t, car_output]
                        == alpha1[car_output][ind - 1]
                        * self.input[t, self.component_options.main_input_carrier]
                        + alpha2[car_output][ind - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_output_on = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_on
                )

                # min part load constraint
                def init_min_partload(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_input_output = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_input_output
        )

        # Bind disjuncts
        def bind_disjunctions(dis, t):
            return [b_tec.dis_input_output[t, i] for i in s_indicators]

        b_tec.disjunction_input_output = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions
        )

        return b_tec

    def _performance_function_type_4(self, b_tec):
        """
        Sets the constraints (input-output and startup/shutdown) for a tec based on tec_type CONV3 with performance
        type 4.

        Type 4 is also a piecewise linear fit to the performance data, based on the number of segments specified. Note
        that this requires a big-m transformation. Again, a minimum part load is possible. Additionally, type 4 includes
        constraints for slow (>1h) startup and shutdown trajectories.

        Based on Equations 9-11, 13 and 15 in Morales-España, G., Ramírez-Elizondo, L., & Hobbs, B. F. (2017). Hidden
        power system inflexibilities imposed by traditional unit commitment formulations. Applied Energy, 191, 223–238.
        https://doi.org/10.1016/J.APENERGY.2017.01.089

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """

        # Transformation required
        self.big_m_transformation_required = 1

        # Performance Parameters
        rated_power = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics
        alpha1 = {}
        alpha2 = {}
        for car in coeff_ti["fit"]:
            bp_x = coeff_ti["fit"][car]["bp_x"]
            alpha1[car] = coeff_ti["fit"][car]["alpha1"]
            alpha2[car] = coeff_ti["fit"][car]["alpha2"]
        min_part_load = coeff_ti["min_part_load"]
        SU_time = dynamics["SU_time"]
        SD_time = dynamics["SD_time"]

        if SU_time <= 0 and SD_time <= 0:
            warn(
                "Having performance_function_type = 4 with no slow SU/SDs usually makes no sense."
            )
        elif SU_time < 0:
            SU_time = 0
        elif SD_time < 0:
            SD_time = 0

        # Calculate SU and SD trajectories
        if SU_time > 0:
            SU_trajectory = []
            for i in range(1, SU_time + 1):
                SU_trajectory.append((min_part_load / (SU_time + 1)) * i)

        if SD_time > 0:
            SD_trajectory = []
            for i in range(1, SD_time + 1):
                SD_trajectory.append((min_part_load / (SD_time + 1)) * i)
            SD_trajectory = sorted(SD_trajectory, reverse=True)

        # slow startups/shutdowns with trajectories
        s_indicators = range(0, SU_time + SD_time + len(bp_x))

        def init_SUSD_trajectories(dis, t, ind):
            if ind == 0:  # technology off
                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                def init_y_off(const, i):
                    if t < len(self.set_t_full) - SU_time or i > SU_time - (
                        len(self.set_t_full) - t
                    ):
                        return b_tec.var_y[t - i + SU_time + 1] == 0
                    else:
                        return (
                            b_tec.var_y[(t - i + SU_time + 1) - len(self.set_t_full)]
                            == 0
                        )

                dis.const_y_off = pyo.Constraint(range(1, SU_time + 1), rule=init_y_off)

                def init_z_off(const, j):
                    if j <= t:
                        return b_tec.var_z[t - j + 1] == 0
                    else:
                        return b_tec.var_z[len(self.set_t_full) + (t - j + 1)] == 0

                dis.const_z_off = pyo.Constraint(range(1, SD_time + 1), rule=init_z_off)

                def init_input_off(const, car_input):
                    return self.input[t, car_input] == 0

                dis.const_input_off = pyo.Constraint(
                    b_tec.set_input_carriers, rule=init_input_off
                )

                def init_output_off(const, car_output):
                    return self.output[t, car_output] == 0

                dis.const_output_off = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_off
                )

            elif ind in range(1, SU_time + 1):  # technology in startup
                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                def init_y_on(const):
                    if t < len(self.set_t_full) - SU_time or ind > SU_time - (
                        len(self.set_t_full) - t
                    ):
                        return b_tec.var_y[t - ind + SU_time + 1] == 1
                    else:
                        return (
                            b_tec.var_y[(t - ind + SU_time + 1) - len(self.set_t_full)]
                            == 1
                        )

                dis.const_y_on = pyo.Constraint(rule=init_y_on)

                def init_z_off(const):
                    if t < len(self.set_t_full) - SU_time or ind > SU_time - (
                        len(self.set_t_full) - t
                    ):
                        return b_tec.var_z[t - ind + SU_time + 1] == 0
                    else:
                        return (
                            b_tec.var_z[(t - ind + SU_time + 1) - len(self.set_t_full)]
                            == 0
                        )

                dis.const_z_off = pyo.Constraint(rule=init_z_off)

                def init_input_SU(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        == b_tec.var_size * SU_trajectory[ind - 1]
                    )

                dis.const_input_SU = pyo.Constraint(rule=init_input_SU)

                def init_output_SU(const, car_output):
                    return (
                        self.output[t, car_output]
                        == alpha1[car_output][0]
                        * self.input[t, self.component_options.main_input_carrier]
                        + alpha2[car_output][0] * b_tec.var_size * rated_power
                    )

                dis.const_output_SU = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_SU
                )

            elif ind in range(
                SU_time + 1, SU_time + SD_time + 1
            ):  # technology in shutdown
                ind_SD = ind - SU_time
                dis.const_x_off = pyo.Constraint(expr=b_tec.var_x[t] == 0)

                def init_z_on(const):
                    if ind_SD <= t:
                        return b_tec.var_z[t - ind_SD + 1] == 1
                    else:
                        return b_tec.var_z[len(self.set_t_full) + (t - ind_SD + 1)] == 1

                dis.const_z_on = pyo.Constraint(rule=init_z_on)

                def init_y_off(const):
                    if ind_SD <= t:
                        return b_tec.var_y[t - ind_SD + 1] == 0
                    else:
                        return b_tec.var_y[len(self.set_t_full) + (t - ind_SD + 1)] == 0

                dis.const_y_off = pyo.Constraint(rule=init_y_off)

                def init_input_SD(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        == b_tec.var_size * SD_trajectory[ind_SD - 1]
                    )

                dis.const_input_SD = pyo.Constraint(rule=init_input_SD)

                def init_output_SD(const, car_output):
                    return (
                        self.output[t, car_output]
                        == alpha1[car_output][0]
                        * self.input[t, self.component_options.main_input_carrier]
                        + alpha2[car_output][0] * b_tec.var_size * rated_power
                    )

                dis.const_output_SD = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_SD
                )

            elif ind > SU_time + SD_time:
                ind_bpx = ind - (SU_time + SD_time)
                dis.const_x_on = pyo.Constraint(expr=b_tec.var_x[t] == 1)

                def init_input_on1(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        >= bp_x[ind_bpx - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_on1 = pyo.Constraint(rule=init_input_on1)

                def init_input_on2(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        <= bp_x[ind_bpx] * b_tec.var_size * rated_power
                    )

                dis.const_input_on2 = pyo.Constraint(rule=init_input_on2)

                def init_output_on(const, car_output):
                    return (
                        self.output[t, car_output]
                        == alpha1[car_output][ind_bpx - 1]
                        * self.input[t, self.component_options.main_input_carrier]
                        + alpha2[car_output][ind_bpx - 1] * b_tec.var_size * rated_power
                    )

                dis.const_input_output_on = pyo.Constraint(
                    b_tec.set_output_carriers, rule=init_output_on
                )

                # min part load relation
                def init_min_partload(const):
                    return (
                        self.input[t, self.component_options.main_input_carrier]
                        >= min_part_load * b_tec.var_size * rated_power
                    )

                dis.const_min_partload = pyo.Constraint(rule=init_min_partload)

        b_tec.dis_SUSD_trajectory = gdp.Disjunct(
            self.set_t_full, s_indicators, rule=init_SUSD_trajectories
        )

        def bind_disjunctions_SUSD(dis, t):
            return [b_tec.dis_SUSD_trajectory[t, k] for k in s_indicators]

        b_tec.disjunction_SUSD_traject = gdp.Disjunction(
            self.set_t_full, rule=bind_disjunctions_SUSD
        )

        return b_tec

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if (
            not self.component_options.performance_function_type == 1
            and "ramping_const_int" in dynamics
            and dynamics["ramping_const_int"] == 1
        ):

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation(const):
                            return (
                                -ramping_rate
                                <= self.input[
                                    t, self.component_options.main_input_carrier
                                ]
                                - self.input[
                                    t - 1, self.component_options.main_input_carrier
                                ]
                            )

                        dis.const_ramping_down_rate = pyo.Constraint(
                            rule=init_ramping_down_rate_operation
                        )

                        def init_ramping_up_rate_operation(const):
                            return (
                                self.input[t, self.component_options.main_input_carrier]
                                - self.input[
                                    t - 1, self.component_options.main_input_carrier
                                ]
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate = pyo.Constraint(
                            rule=init_ramping_up_rate_operation
                        )

                    elif ind == 1:  # startup, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:
            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                set_t_rr = self.set_t_performance
            else:
                if (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 1
                ):
                    sequence = data["k_means_specs"]["sequence"]
                elif (
                    data["config"]["optimization"]["typicaldays"]["method"]["value"]
                    == 2
                ):
                    sequence = self.sequence

                # init bounds at full res
                bounds_rr_full = {
                    "input": self.fitting_class.calculate_input_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    )
                }

                # create input variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )

                b_tec.const_link_full_resolution_rr = link_full_resolution_to_clustered(
                    self.input,
                    b_tec.var_input_rr_full,
                    self.set_t_full,
                    sequence,
                    b_tec.set_input_carriers,
                )

                input_aux_rr = b_tec.var_input_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate(const, t):
                if t > 1:
                    return (
                        -ramping_rate
                        <= input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate
            )

            def init_ramping_up_rate(const, t):
                if t > 1:
                    return (
                        input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate
            )

        return b_tec



import logging
from pyomo.core import Constraint, value


def get_infeasible_constraints(m, tolerance=1e-3):
    """
    Gets violated constraints of a pyomo model and send them to the logger

    :param m: pyomo model
    :param Path save_path: path to save
    :param tolerance: tolerance of constraint violation
    :return: None
    """
    logger = logging.getLogger(__name__)

    for constr in m.component_data_objects(
        ctype=Constraint, active=True, descend_into=True
    ):
        body_value = value(constr.body, exception=False)
        infeasible = 0
        infeasibility = 0
        if body_value is not None:
            if constr.has_lb():
                lb = value(constr.lower, exception=False)
                if lb is None:
                    infeasible = 4 | 1
                elif lb - body_value > tolerance:
                    infeasible = 1
                    infeasibility = infeasibility + (lb - body_value)
            if constr.has_ub():
                ub = value(constr.upper, exception=False)
                if ub is None:
                    infeasible |= 4 | 2
                elif body_value - ub > tolerance:
                    infeasible = 1
                    infeasibility = infeasibility + (body_value - ub)

        if infeasible:
            logger.info(constr.name + " is infeasible by " + str(infeasibility))



import pyomo.environ as pyo
import pyomo.gdp as gdp
import numpy as np
import pandas as pd

from ...utilities import get_attribute_from_dict, link_full_resolution_to_clustered
from ..technology import Technology


class HydroOpen(Technology):
    """
    Open pumped hydro technology

    Resembles a pumped hydro plant with additional natural inflows (defined in
    climate data). Note that this technology only works for one carrier, and thus the
    carrier index is dropped in the below notation.

    **Variable declarations:**

    - Storage level in :math:`t`: :math:`E_t`

    - Charging in :math:`t`: :math:`Input_{t}`

    - Discharging in :math:`t`: :math:`Output_{t}`

    **Constraint declarations:**

    The following constants are used:

    - :math:`{\\eta}_{in}`: Charging efficiency

    - :math:`{\\eta}_{out}`: Discharging efficiency

    - :math:`{\\lambda}`: Self-Discharging coefficient

    - :math:`Input_{max}`: Maximal charging capacity in one time-slice

    - :math:`Output_{max}`: Maximal discharging capacity in one time-slice

    - :math:`Natural_Inflow{t}`: Natural water inflow in time slice (can be negative, i.e. being an outflow)


    - Maximal charging and discharging:

      .. math::
        Input_{t} \\leq Input_{max}

      .. math::
        Output_{t} \\leq Output_{max}

    - Size constraint:

      .. math::
        E_{t} \\leq S

    - Storage level calculation:

      .. math::
        E_{t} = E_{t-1} * (1 - \\lambda) + {\\eta}_{in} * Input_{t} - 1 / {\\eta}_{out} * Output_{t} + Natural_Inflow_{t}

    - If ``allow_only_one_direction == 1``, then only input or output can be unequal to zero in each respective time
      step (otherwise, simultanous charging and discharging can lead to unwanted 'waste' of energy/material).

    - Additionally, ramping rates of the technology can be constraint (for input and
      output).

      .. math::
         -rampingrate \\leq Input_{t, maincar} - Input_{t-1, maincar} \\leq rampingrate

      .. math::
         -rampingrate \\leq \\sum(Input_{t, car}) - \\sum(Input_{t-1, car}) \\leq rampingrate

    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits technology performance

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(HydroOpen, self).fit_technology_performance(climate_data, location)

        # Coefficients
        for par in self.input_parameters.performance_data["performance"]:
            self.processed_coeff.time_independent[par] = (
                self.input_parameters.performance_data["performance"][par]
            )

        # Natural inflow
        if self.name + "_inflow" in climate_data:
            self.processed_coeff.time_dependent_full["hydro_inflow"] = climate_data[
                self.name + "_inflow"
            ]
        else:
            raise Exception(
                "Using Technology Type Hydro_Open requires a hydro_natural_inflow in climate data"
                " to be defined for this node. Add a column in the climate data for respective node with column name"
                f" {self.name}_inflow"
            )

        # Maximum discharge
        if self.input_parameters.performance_data["maximum_discharge_time_discrete"]:
            if self.name + "_maximum_discharge" in climate_data:
                self.processed_coeff.time_dependent_full["hydro_maximum_discharge"] = (
                    climate_data[self.name + "_maximum_discharge"]
                )
            else:
                raise Exception(
                    "Using Technology Type Hydro_Open with maximum_discharge_time_discrete == 1 requires "
                    "hydro_maximum_discharge to be defined for this node."
                )

        # Options
        self.component_options.other["allow_only_one_direction"] = (
            get_attribute_from_dict(
                self.input_parameters.performance_data, "allow_only_one_direction", 0
            )
        )
        self.component_options.other["can_pump"] = get_attribute_from_dict(
            self.input_parameters.performance_data, "can_pump", 1
        )
        self.component_options.other["allow_only_one_direction_precise"] = (
            get_attribute_from_dict(
                self.input_parameters.performance_data,
                "allow_only_one_direction_precise",
                1,
            )
        )
        self.component_options.other["maximum_discharge_time_discrete"] = (
            get_attribute_from_dict(
                self.input_parameters.performance_data,
                "maximum_discharge_time_discrete",
                1,
            )
        )

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(HydroOpen, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Output Bounds
        for car in self.component_options.output_carrier:
            self.bounds["output"][car] = np.column_stack(
                (
                    np.zeros(shape=(time_steps)),
                    np.ones(shape=(time_steps))
                    * self.input_parameters.performance_data["performance"][
                        "discharge_max"
                    ],
                )
            )

        # Input Bounds
        for car in self.component_options.input_carrier:
            self.bounds["input"][car] = np.column_stack(
                (
                    np.zeros(shape=(time_steps)),
                    np.ones(shape=(time_steps))
                    * self.input_parameters.performance_data["performance"][
                        "charge_max"
                    ],
                )
            )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type Hydro_Open

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(HydroOpen, self).construct_tech_model(
            b_tec, data, set_t_full, set_t_clustered
        )

        config = data["config"]

        # DATA OF TECHNOLOGY
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics
        allow_only_one_direction = self.component_options.other[
            "allow_only_one_direction"
        ]

        eta_in = coeff_ti["eta_in"]
        eta_out = coeff_ti["eta_out"]
        eta_lambda = coeff_ti["lambda"]
        charge_max = coeff_ti["charge_max"]
        discharge_max = coeff_ti["discharge_max"]
        spilling_max = coeff_ti["spilling_max"]
        hydro_natural_inflow = coeff_td["hydro_inflow"]

        if config["optimization"]["timestaging"]["value"] != 0:
            nr_timesteps_averaged = config["optimization"]["timestaging"]["value"]
        else:
            nr_timesteps_averaged = 1

        # Additional decision variables
        b_tec.var_storage_level = pyo.Var(
            self.set_t_performance,
            b_tec.set_input_carriers,
            domain=pyo.NonNegativeReals,
            bounds=(b_tec.para_size_min, b_tec.para_size_max),
        )
        b_tec.var_spilling = pyo.Var(
            self.set_t_performance,
            domain=pyo.NonNegativeReals,
            bounds=(b_tec.para_size_min, b_tec.para_size_max),
        )

        # Abdditional parameters

        # Size constraint
        def init_size_constraint(const, t, car):
            return b_tec.var_storage_level[t, car] <= b_tec.var_size

        b_tec.const_size = pyo.Constraint(
            self.set_t_performance, b_tec.set_input_carriers, rule=init_size_constraint
        )

        # Storage level calculation
        def init_storage_level(const, t, car):
            if t == 1:  # couple first and last time interval
                return (
                    b_tec.var_storage_level[t, car]
                    == b_tec.var_storage_level[max(self.set_t_performance), car]
                    * (1 - eta_lambda) ** nr_timesteps_averaged
                    + (
                        eta_in * self.input[t, car]
                        - 1 / eta_out * self.output[t, car]
                        - b_tec.var_spilling[t]
                    )
                    * sum(
                        (1 - eta_lambda) ** i for i in range(0, nr_timesteps_averaged)
                    )
                    + hydro_natural_inflow.iloc[t - 1]
                )
            else:  # all other time intervals
                return (
                    b_tec.var_storage_level[t, car]
                    == b_tec.var_storage_level[t - 1, car]
                    * (1 - eta_lambda) ** nr_timesteps_averaged
                    + (
                        eta_in * self.input[t, car]
                        - 1 / eta_out * self.output[t, car]
                        - b_tec.var_spilling[t]
                    )
                    * sum(
                        (1 - eta_lambda) ** i for i in range(0, nr_timesteps_averaged)
                    )
                    + hydro_natural_inflow.iloc[t - 1]
                )

        b_tec.const_storage_level = pyo.Constraint(
            self.set_t_performance, b_tec.set_input_carriers, rule=init_storage_level
        )

        if not self.component_options.other["can_pump"]:

            def init_input_zero(const, t, car):
                return self.input[t, car] == 0

            b_tec.const_input_zero = pyo.Constraint(
                self.set_t_performance, b_tec.set_input_carriers, rule=init_input_zero
            )

        # This makes sure that only either input or output is larger zero.
        if allow_only_one_direction == 1:

            # Cut according to Germans work
            def init_cut_bidirectional(const, t, car):
                return (
                    self.output[t, car] / discharge_max
                    + self.input[t, car] / charge_max
                    <= b_tec.var_size
                )

            b_tec.const_cut_bidirectional = pyo.Constraint(
                self.set_t_performance,
                b_tec.set_input_carriers,
                rule=init_cut_bidirectional,
            )

            # Disjunct modelling
            if self.component_options.other["allow_only_one_direction_precise"]:
                self.big_m_transformation_required = 1
                s_indicators = range(0, 2)

                def init_input_output(dis, t, ind):
                    if ind == 0:  # input only

                        def init_output_to_zero(const, car_input):
                            return self.output[t, car_input] == 0

                        dis.const_output_to_zero = pyo.Constraint(
                            b_tec.set_input_carriers, rule=init_output_to_zero
                        )

                    elif ind == 1:  # output only

                        def init_input_to_zero(const, car_input):
                            return self.input[t, car_input] == 0

                        dis.const_input_to_zero = pyo.Constraint(
                            b_tec.set_input_carriers, rule=init_input_to_zero
                        )

                b_tec.dis_input_output = gdp.Disjunct(
                    self.set_t_performance, s_indicators, rule=init_input_output
                )

                # Bind disjuncts
                def bind_disjunctions(dis, t):
                    return [b_tec.dis_input_output[t, i] for i in s_indicators]

                b_tec.disjunction_input_output = gdp.Disjunction(
                    self.set_t_performance, rule=bind_disjunctions
                )

        # Maximal charging and discharging rates
        def init_maximal_charge(const, t, car):
            return self.input[t, car] <= charge_max * b_tec.var_size

        b_tec.const_max_charge = pyo.Constraint(
            self.set_t_performance, b_tec.set_input_carriers, rule=init_maximal_charge
        )

        def init_maximal_discharge(const, t, car):
            return self.output[t, car] <= discharge_max * b_tec.var_size

        b_tec.const_max_discharge = pyo.Constraint(
            self.set_t_performance,
            b_tec.set_input_carriers,
            rule=init_maximal_discharge,
        )

        if self.component_options.other["maximum_discharge_time_discrete"]:

            def init_maximal_discharge2(const, t, car):
                return self.output[t, car] <= coeff_td["hydro_maximum_discharge"][t - 1]

            b_tec.const_max_discharge2 = pyo.Constraint(
                self.set_t_performance,
                b_tec.set_input_carriers,
                rule=init_maximal_discharge2,
            )

        # Maximum spilling
        def init_maximal_spilling(const, t):
            return b_tec.var_spilling[t] <= spilling_max * b_tec.var_size

        b_tec.const_max_spilling = pyo.Constraint(
            self.set_t_performance, rule=init_maximal_spilling
        )

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data)

        return b_tec

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report technology operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        super(HydroOpen, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "spilling",
            data=[model_block.var_spilling[t].value for t in self.set_t_performance],
        )
        for car in model_block.set_input_carriers:
            h5_group.create_dataset(
                "storage_level_" + car,
                data=[
                    model_block.var_storage_level[t, car].value
                    for t in self.set_t_performance
                ],
            )

    def _define_ramping_rates(self, b_tec, data):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if (
            "ramping_const_int" in self.performance_data
            and self.performance_data["ramping_const_int"] == 1
        ):

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation_in(const):
                            return -ramping_rate <= sum(
                                self.input[t, car_input] - self.input[t - 1, car_input]
                                for car_input in b_tec.set_input_carriers
                            )

                        dis.const_ramping_down_rate_in = pyo.Constraint(
                            rule=init_ramping_down_rate_operation_in
                        )

                        def init_ramping_up_rate_operation_in(const):
                            return (
                                sum(
                                    self.input[t, car_input]
                                    - self.input[t - 1, car_input]
                                    for car_input in b_tec.set_input_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate_in = pyo.Constraint(
                            rule=init_ramping_up_rate_operation_in
                        )

                        def init_ramping_down_rate_operation_out(const):
                            return -ramping_rate <= sum(
                                self.output[t, car_output]
                                - self.output[t - 1, car_output]
                                for car_output in b_tec.set_output_carriers
                            )

                        dis.const_ramping_down_rate_out = pyo.Constraint(
                            rule=init_ramping_down_rate_operation_out
                        )

                        def init_ramping_up_rate_operation_out(const):
                            return (
                                sum(
                                    self.output[t, car_output]
                                    - self.output[t - 1, car_output]
                                    for car_output in b_tec.set_output_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate_out = pyo.Constraint(
                            rule=init_ramping_up_rate_operation_out
                        )

                    elif ind == 1:  # startup, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:
            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                output_aux_rr = self.output
                set_t_rr = self.set_t_performance
            else:
                # init bounds at full res
                bounds_rr_full = {
                    "input": self.fitting_class.calculate_input_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    ),
                    "output": self.fitting_class.calculate_output_bounds(
                        self.component_options.size_based_on, len(self.set_t_full)
                    ),
                }

                # create input and output variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                def init_output_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["output"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )
                b_tec.var_output_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_output_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_output_bounds,
                )

                b_tec.const_link_full_resolution_rr_input = (
                    link_full_resolution_to_clustered(
                        self.input,
                        b_tec.var_input_rr_full,
                        self.set_t_full,
                        self.sequence,
                        b_tec.set_input_carriers,
                    )
                )

                b_tec.const_link_full_resolution_rr_output = (
                    link_full_resolution_to_clustered(
                        self.output,
                        b_tec.var_output_rr_full,
                        self.set_t_full,
                        sequence_storage,
                        b_tec.set_output_carriers,
                    )
                )

                input_aux_rr = b_tec.var_input_rr_full
                output_aux_rr = b_tec.var_output_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate_input(const, t):
                # -rampingRate <= input[t] - input[t-1]
                if t > 1:
                    return (
                        -ramping_rate
                        <= input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate_input
            )

            def init_ramping_up_rate_input(const, t):
                # input[t] - input[t-1] <= rampingRate
                if t > 1:
                    return (
                        input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate_input
            )

            def init_ramping_down_rate_output(const, t):
                # -rampingRate <= output[t] - output[t-1]
                if t > 1:
                    return -ramping_rate <= sum(
                        output_aux_rr[t, car_output] - output_aux_rr[t - 1, car_output]
                        for car_output in b_tec.set_output_carriers
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate_output = pyo.Constraint(
                self.set_t_performance, rule=init_ramping_down_rate_output
            )

            def init_ramping_down_rate_output(const, t):
                # output[t] - output[t-1] <= rampingRate
                if t > 1:
                    return (
                        sum(
                            output_aux_rr[t, car_output]
                            - output_aux_rr[t - 1, car_output]
                            for car_output in b_tec.set_output_carriers
                        )
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate_output = pyo.Constraint(
                self.set_t_performance, rule=init_ramping_down_rate_output
            )

        return b_tec
